from tqdm import tqdm
import torch
ch = torch
import poisoning_adversary
import numpy as np
import os
import poison_profiles
from argparse import ArgumentParser
from torchvision import transforms
from copy import deepcopy
import poisoned_datasets
import torchvision
from robustness import model_utils, attacker, attack_steps, defaults
from robustness.train import train_model, _model_loop
import gammas
import selection_strategy
import torch.nn.functional as F
import matplotlib.pyplot as plt


CLASS_SIZE = 32
# RESOLUTION = 1000
DIMENSIONS=16
STEP_SIZE=0.01
EPS=0.2
ITERATIONS=25
EPOCHS = 1500
LR = 0.001

def unzip(l):
    return [i[0] for i in l], [i[1] for i in l]

'''
Dataset
'''
def get_dataset(dist, poison_cnt):
    x = np.concatenate([np.random.normal(-10.0, 1.0, size=[CLASS_SIZE-poison_cnt, DIMENSIONS]),
         np.random.normal(10.0, 1.0, size=[CLASS_SIZE-poison_cnt, DIMENSIONS])], axis=0)
    y = [int(i >= len(x)//2) for i, _ in enumerate(x)]
    px = np.random.normal(size=[poison_cnt*2, DIMENSIONS])
    norm = (np.sum(px ** 2, axis=-1) ** 0.5).reshape([-1, 1])
    px = dist * px / norm
    offset = np.asarray([-10.0 for _ in range(poison_cnt)] + [10.0 for _ in range(poison_cnt)]).reshape([-1, 1])
    px = px + offset
    py = [2 for i, _ in enumerate(px)]
    x = np.concatenate([x, px]).astype(np.float32)
    y = np.concatenate([y, py])
    train = list(zip(x, y))
    
    x = np.concatenate([np.random.normal(-10.0, 1.0, size=[CLASS_SIZE, DIMENSIONS]),
         np.random.normal(10.0, 1.0, size=[CLASS_SIZE, DIMENSIONS])], axis=0).astype(np.float32)
    y = [int(i >= len(x)//2) for i, _ in enumerate(x)]
    test = list(zip(x, y))
    
    return train, test

from robustness.tools import helpers
from robustness import attack_steps

class LinfStep(attack_steps.AttackerStep):
    """
    Attack step for :math:`\ell_\infty` threat model. Given :math:`x_0`
    and :math:`\epsilon`, the constraint set is given by:
    .. math:: S = \{x | \|x - x_0\|_\infty \leq \epsilon\}
    """
    def project(self, x):
        """
        """
        diff = x - self.orig_input
        diff = ch.clamp(diff, -self.eps, self.eps)
        return self.orig_input + diff

    def step(self, x, g):
        """
        """
        step = ch.sign(g) * self.step_size
        return x + step

    def random_perturb(self, x):
        """
        """
        new_x = x + 2 * (ch.rand_like(x) - 0.5) * self.eps
        return new_x

class Attacker(ch.nn.Module):
    """
    Attacker class, used to make adversarial examples.

    This is primarily an internal class, you probably want to be looking at
    :class:`robustness.attacker.AttackerModel`, which is how models are actually
    served (AttackerModel uses this Attacker class).

    However, the :meth:`robustness.Attacker.forward` function below
    documents the arguments supported for adversarial attacks specifically.
    """
    def __init__(self, model):
        """
        Initialize the Attacker

        Args:
            nn.Module model : the PyTorch model to attack
            Dataset dataset : dataset the model is trained on, only used to get mean and std for normalization
        """
        super(Attacker, self).__init__()
        self.model = model

    def forward(self, x, target, *_, constraint, eps, step_size, iterations,
                random_start=True, random_restarts=False, do_tqdm=False,
                targeted=False, custom_loss=None, should_normalize=False,
                orig_input=None, use_best=True, return_image=True,
                est_grad=None, mixed_precision=False):
        # Can provide a different input to make the feasible set around
        # instead of the initial point
        if orig_input is None: orig_input = x.detach()
        orig_input = orig_input

        # Multiplier for gradient ascent [untargeted] or descent [targeted]
        m = -1 if targeted else 1

        # Initialize step class and attacker criterion
        criterion = ch.nn.CrossEntropyLoss(reduction='none')
        step_class = LinfStep
        step = step_class(eps=eps, orig_input=orig_input, step_size=step_size) 

        def calc_loss(inp, target):
            '''
            Calculates the loss of an input with respect to target labels
            Uses custom loss (if provided) otherwise the criterion
            '''
            if should_normalize:
                inp = self.normalize(inp)
            output = self.model(inp)
            if custom_loss:
                return custom_loss(self.model, inp, target)

            return criterion(output, target), output

        # Main function for making adversarial examples
        def get_adv_examples(x):
#             print('start', x)
            # Random start (to escape certain types of gradient masking)
            if random_start:
                x = step.random_perturb(x)

#             print('random per', x)
            
            iterator = range(iterations)
            if do_tqdm: iterator = tqdm(iterator)

            # Keep track of the "best" (worst-case) loss and its
            # corresponding input
            best_loss = None
            best_x = None

            # A function that updates the best loss and best input
            def replace_best(loss, bloss, x, bx):
                if bloss is None:
                    bx = x.clone().detach()
                    bloss = loss.clone().detach()
                else:
                    replace = m * bloss < m * loss
                    bx[replace] = x[replace].clone().detach()
                    bloss[replace] = loss[replace]

                return bloss, bx

            # PGD iterates
            for _ in iterator:
                x = x.clone().detach().requires_grad_(True)
                losses, out = calc_loss(step.to_image(x), target)
                assert losses.shape[0] == x.shape[0], \
                        'Shape of losses must match input!'

                loss = ch.mean(losses)

                if step.use_grad:
                    if (est_grad is None) and mixed_precision:
                        with amp.scale_loss(loss, []) as sl:
                            sl.backward()
                        grad = x.grad.detach()
                        x.grad.zero_()
                    elif (est_grad is None):
                        grad, = ch.autograd.grad(m * loss, [x])
                    else:
                        f = lambda _x, _y: m * calc_loss(step.to_image(_x), _y)[0]
                        grad = helpers.calc_est_grad(f, x, target, *est_grad)
                else:
                    grad = None

                with ch.no_grad():
                    args = [losses, best_loss, x, best_x]
                    best_loss, best_x = replace_best(*args) if use_best else (losses, x)

                    x = step.step(x, grad)
                    x = step.project(x)
                    if do_tqdm: iterator.set_description("Current loss: {l}".format(l=loss))

            # Save computation (don't compute last loss) if not use_best
            if not use_best: 
                ret = x.clone().detach()
                return step.to_image(ret) if return_image else ret

            losses, _ = calc_loss(step.to_image(x), target)
            args = [losses, best_loss, x, best_x]
            best_loss, best_x = replace_best(*args)
            return step.to_image(best_x) if return_image else best_x

        # Random restarts: repeat the attack and find the worst-case
        # example for each input in the batch
        if random_restarts:
            to_ret = None

            orig_cpy = x.clone().detach()
            for _ in range(random_restarts):
                adv = get_adv_examples(orig_cpy)

                if to_ret is None:
                    to_ret = adv.detach()

                _, output = calc_loss(adv, target)
                corr, = helpers.accuracy(output, target, topk=(1,), exact=True)
                corr = corr.byte()
                misclass = ~corr
                to_ret[misclass] = adv[misclass]

            adv_ret = to_ret
        else:
            adv_ret = get_adv_examples(x)

        return adv_ret

class AttackerModel(ch.nn.Module):
    def __init__(self, model):
        super(AttackerModel, self).__init__()
#         self.normalizer = helpers.InputNormalize(dataset.mean, dataset.std)
        self.model = model
        self.attacker = Attacker(model)

    def forward(self, inp, target=None, make_adv=True,
                with_image=True, **attacker_kwargs):
        
        if make_adv:
            assert target is not None
            prev_training = bool(self.training)
            self.eval()
            adv = self.attacker(inp, target, **attacker_kwargs)
            if prev_training:
                self.train()

            inp = adv

        normalized_inp = inp

        output = self.model(normalized_inp)
        if with_image:
            return (output, inp)
        return output


def get_model(units=128, inputs=DIMENSIONS):
    return torch.nn.Sequential(
            torch.nn.Linear(inputs, units),
            torch.nn.ReLU(),
            torch.nn.Linear(units, 4*units),
            torch.nn.ReLU(),
            torch.nn.Linear(4*units, 4*units),
            torch.nn.ReLU(),
            torch.nn.Linear(4*units, units),
            torch.nn.ReLU(),
            torch.nn.Linear(units, 3)
        )

def test(model, data_loader):
    results = []
    for x, y in data_loader:
        output = torch.argmax(model(x), dim=-1)
        results.extend((output.view(-1) == y).detach().numpy())
    return results.count(True) / len(results)

def adv(model, data_loader):
    amodel = AttackerModel(model)
    res = []
    for x, y in data_loader:
        _, input = amodel(x, y, make_adv=True,
             constraint='inf', step_size=STEP_SIZE,
             eps=EPS, iterations=ITERATIONS, random_restarts=1,
                      targeted=False)
#         print(input)
        out = model(input)
        out = torch.argmax(out, dim=-1) == y
        res.extend(out.detach().numpy())
#     print(res)
    return res.count(True) / len(res)


def train(model, data_loader, test_loader):
    amodel = AttackerModel(model)
    epochs=EPOCHS
    lr=LR
    optimizer = torch.optim.Adam(
        params=model.parameters(),
        lr=lr,
    )
    loss = torch.nn.CrossEntropyLoss()
    loss_log = []
    test_log = [test(model, data_loader)]
    for epoch in tqdm(range(epochs)):
        loss_avg = 0.0
        for x, y in data_loader:
            optimizer.zero_grad()
            
            _, input = amodel(x, y, make_adv=True,
            constraint='inf', step_size=STEP_SIZE,
            eps=EPS, iterations=ITERATIONS, random_restarts=1, targeted=False)
            
            x = input
            
            output = model(x)
            loss_val = loss(output, target=y)
            loss_avg = loss_avg * 0.5 + 0.5 * loss_val.cpu().detach().numpy()
            loss_val.backward()
            optimizer.step()
        test_log.append(test(model, data_loader))
        loss_log.append(loss_avg)
    train_acc = test(model, data_loader)
    if train_acc < 1.0:
        print(train_acc)
    
    return loss_log, test_log, test(model, test_loader), adv(model, test_loader), train_acc


def run(seed, r, return_net=False):
    np.random.seed(seed)
    torch.manual_seed(seed)
    model = get_model()
    train_, test_ = get_dataset(r, 2)
    train_ = torch.utils.data.DataLoader(train_, batch_size=8)
    test_ = torch.utils.data.DataLoader(test_, batch_size=8)
    loss, train_acc, test_acc, adv_acc, model = train(model, train_, test_)
    if return_net is False:
        return test_acc, adv_acc
    else:
        return test_acc, adv_acc, model

seeds = range(10)

import sys

r = float(sys.argv[1])
print(r, type(r))

# rs = [0.4, 0.8, 1.2, 1.6, 2.0, 2.4, 2.8, 3.2, 3.6, 4.0, 4.4, 4.8, 5.2, 5.6, 6.0]
# rs = [0.4, 0.8, 1.6, 2.4, 3.2, 4.0, 5.2, 6.0]
from itertools import product

nat_eval, adv_eval, r_ = [], [], []
for seed in seeds:
    n, a = run(seed, r)
    nat_eval.append(n)
    adv_eval.append(a)
    r_.append(r)
# run(100, 1.0)
import pandas as pd
df = pd.DataFrame(data={'nat': nat_eval, 'adv': adv_eval, 'r': r_})
print(df)
print(nat_eval)
print(adv_eval)
print(r_)

df.to_pickle(os.path.join(os.environ['SHARDATA'], 'pgd-training-gaussian', 'r-' + str(r) + '.pkl'))
